/**
 * Program entrypoint and intiialization.
 */


// System Import Headers.
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <iterator>
#include <string>
#include <vector>

// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "image.h"
    #include "utility.h"
#endif

#ifndef user_headers_logging
    #define user_headers_logging
    #include "logging.h"
#endif


// Method Declaration.
void process_dir(std::string path_str);
void process_file(std::string path_str);


// Global Variables.
#ifndef LOGGING_CLASS_INIT
    #define LOGGING_CLASS_INIT
    Logging logger;
#endif


/*
 * Program's main. Initializes and runs program.
 */
int main(int argc, char* argv[]) {
    printf("Starting program.\n");
    printf("\n");
    printf("\n");

    // Initialize logging and clear any existing log contents for new program run.
    logger.clear_log_files();

    // Check provided number of args.
    if (argc < 2) {
        // No args provided.
        printf("Provided too few program args. Expected one arg.");
    } else if (argc > 2) {
        // More than one arg provided.
        printf("Provided too many program args. Expected one arg.");
    } else {
        // Correct number of args provided. Pass to image class for processing.

        // Check data type.
        std::string path_str = argv[1];
        path_validator_struct* return_struct = validate_path(path_str);
        int path_type = return_struct->file_type;
        if (return_struct->err_code) {
            std::cerr << "Error verifying file type: " << return_struct->err_code.message() << std::endl;
        }
        free_path_validator_struct(return_struct);

        // Handle based on provided type.
        if (path_type == TYPE_DIR) {
            // Process dir.
            process_dir(path_str);
        } else if (path_type == TYPE_FILE) {
            // Process file.
            process_file(path_str);
        }
    }

    printf("\n");
    printf("\n");
    printf("Terminating program.\n");
    exit(0);
}


/**
 * Finds and processes all files in directory.
 * Assumes directory is full of image files.
 */
void process_dir(std::string path_str) {
    logger.debug("process_dir():");
    logger.debug("    path_str: " + path_str);
    logger.debug("");

    logger.timer("");
    logger.timer("");
    logger.timer("");
    logger.timer("Handling dir \"" + path_str + "\"");

    // Track time to process.
    time_t start_time = time(0);

    // Process all files in directory.
    for (const auto & entry : std::filesystem::directory_iterator(path_str)) {
        // Get as string.
        std::string file_path_str{entry.path().u8string()};
        // std::cout << file_path_str << std::endl;

        // Check if new path value is directory or file.
        path_validator_struct* return_struct = validate_path(file_path_str);
        int path_type = return_struct->file_type;
        if (return_struct->err_code) {
            std::cerr << "Error verifying file type: " << return_struct->err_code.message() << std::endl;
        }
        free_path_validator_struct(return_struct);

        // Handle based on provided type.
        if (path_type == TYPE_DIR) {
            // Process dir.
            process_dir(file_path_str);
        } else if (path_type == TYPE_FILE) {
            // Process file.
            process_file(file_path_str);
        }
    }

    // Calculate time processing dir.
    double total_time = difftime( time(0), start_time);
    logger.timer("");
    logger.timer("Time to Process Dir \"" + path_str + "\": ");
    double seconds_elapsed = fmod(total_time, 60);
    double minutes_elapsed = (total_time - seconds_elapsed) / 60;
    std::ostringstream second_format_string;
    std::ostringstream minute_format_string;
    second_format_string << std::setprecision(0) << std::fixed << seconds_elapsed;
    minute_format_string << std::setprecision(0) << std::fixed << minutes_elapsed;
    logger.timer(minute_format_string.str() + " min, " + second_format_string.str() + " seconds");
}


/**
 * Processes a single image file.
 */
void process_file(std::string path_str) {
    logger.debug("process_file():");
    logger.debug("    path_str: " + path_str);
    logger.debug("");

    logger.info("");
    logger.info("");
    logger.timer("");
    logger.timer("Handling file \"" + path_str + "\"");

    // Track time to process.
    time_t start_time = time(0);

    // Import image data to class.
    Image image(path_str);
    image.display_properties();

    // Manipulate and save image data.
    image.compute();
    // image.save();

    // Calculate time processing file.
    double total_time = difftime( time(0), start_time);
    logger.timer("Time to Process File \"" + path_str + "\": ");
    double seconds_elapsed = fmod(total_time, 60);
    double minutes_elapsed = (total_time - seconds_elapsed) / 60;
    std::ostringstream second_format_string;
    std::ostringstream minute_format_string;
    second_format_string << std::setprecision(0) << std::fixed << seconds_elapsed;
    minute_format_string << std::setprecision(0) << std::fixed << minutes_elapsed;
    logger.timer(minute_format_string.str() + " min, " + second_format_string.str() + " seconds");
}
